//
// -- LICENSE: The MIT License(MIT)
//
// Copyright(c) 2021 DinS - https://dins.site
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
#include "MpgVideo.h"
#define PL_MPEG_IMPLEMENTATION
#include "pl_mpeg.h"
#include <SDL/SDL_audio.h>
#include <vector>


MpgVideo::MpgVideo(Context* context) : Resource(context)
{
}

MpgVideo::~MpgVideo()
{
    plm_destroy(_pMpg);
}

void MpgVideo::RegisterObject(Context* context)
{
    context->RegisterFactory<MpgVideo>("Resource");
}

//不使用跟Urho3D有关的内容，避免线程冲突
bool MpgVideo::BeginLoad(Deserializer& source)
{
    //1.读取文件内容(sizeof(uint8) == 1)
    unsigned nDataSize = source.GetSize();
    uint8_t* pBuffer = (uint8_t *)malloc(nDataSize);
    source.Read(pBuffer, nDataSize);
    //2.解析基本信息
    //plm_t* pMpg = plm_create_with_filename(strPath.c_str());
    _pMpg = plm_create_with_memory(pBuffer, nDataSize, true);     //传true则plmpeg销毁时会清空buffer
    if (_pMpg == nullptr)
        return false;
    _nWidth = plm_get_width(_pMpg);
    _nHeight = plm_get_height(_pMpg);
    _dFps = plm_get_framerate(_pMpg);        //对于mpg应该约等于30
    _nFreq = plm_get_samplerate(_pMpg);     //对于mpg应该都是48000
    //double dDuration = plm_get_duration(_pMpg);
    //3.解析音轨(fast)
    std::vector<float> vecMpgAudioStream;       //因为事前不知道有多长，所以用vector省事
    plm_samples_t* pSample = plm_decode_audio(_pMpg);        //每调用一次就自动推进到下一帧，如果全部解析完或错误或没有音轨则返回nullptr
    while (pSample != nullptr)
    {
        //将解析出来的音频拷贝出来，长度是两倍的count因为是双声道，每个sample是normalized (-1, 1) float
        vecMpgAudioStream.insert(vecMpgAudioStream.end(), pSample->interleaved, pSample->interleaved + pSample->count * 2);
        pSample = plm_decode_audio(_pMpg);
    }
    //4.调整音频格式（Urho3D打开的SDL音频格式是AUDIO_S16，mpg解析出来的是AUDIO_F32）
    //https://wiki.libsdl.org/SDL_ConvertAudio
    SDL_AudioCVT cvt;
    SDL_BuildAudioCVT(&cvt, AUDIO_F32, 2, _nFreq, AUDIO_S16, 2, _nFreq);
    cvt.len = sizeof(float) * vecMpgAudioStream.size();        //source音频长度(in byte)
    cvt.buf = (Uint8 *) SDL_malloc(cvt.len * cvt.len_mult);     //分配足够的buffer
    memcpy(cvt.buf, vecMpgAudioStream.data(), cvt.len);        //把source音频拷贝到buffer
    SDL_ConvertAudio(&cvt);             //返回0失败，不过这里并不需要检查，允许失败
    //保存调整后的音频
    _pRawAudio = cvt.buf;
    _nRawAudioLen = cvt.len_cvt;
    //5.解析图像采取Just-in-Time方式

    return true;
}

//主线程里格式化为Urho3D资源
bool MpgVideo::EndLoad()
{
    //将raw audio stream(pcm)格式化为Sound
    SharedPtr<Sound> pSound(new Sound(GetContext()));
    pSound->SetData(_pRawAudio, _nRawAudioLen);          //内部会memcpy
    pSound->SetFormat(_nFreq, true, true);
    _pAudioStream = pSound;
     SDL_free(_pRawAudio);
    _pRawAudio = nullptr;
    return true;
}

SharedPtr<Sprite2D> MpgVideo::GetNextFrame()
{
    //0.解析原始像素
    //每调用一次解析出一帧(YCrCb格式)，再次调用会解析出下一帧，如果全部解析完或错误或本来就没有图像则返回nullptr
    plm_frame_t* pFrame = plm_decode_video(_pMpg);
    if (pFrame == nullptr)
        return nullptr;
    uint8_t *rgb_buffer = (uint8_t *)malloc(_nWidth * _nHeight * 3);        //分配内存装载rgb buffer，大小是帧高*帧宽*每个像素有RGB三个值
    plm_frame_to_rgb(pFrame, rgb_buffer, _nWidth * 3);       //转换为RGB格式(最后一个参数是stride), slow!
    //1.将raw pixel转变为Image
    SharedPtr<Image> pImg(new Image(GetContext()));
    pImg->SetSize(_nWidth, _nHeight, 3);     //帧高*帧宽*RGB
    pImg->SetData(rgb_buffer);        //内部会memcpy，长度由SetSize决定（可能很耗时）
    free(rgb_buffer);
    //2.构造Texture2D
    SharedPtr<Texture2D> pTexture(new Texture2D(GetContext()));
    pTexture->SetData(pImg);        //不使用alpha
    //3.构造Sprite2D
    SharedPtr<Sprite2D> pSp(new Sprite2D(GetContext()));
    pSp->SetTexture(pTexture);
    pSp->SetRectangle(Urho3D::IntRect(0, 0, _nWidth, _nHeight));        //必须手动设置不然是0不显示任何内容
    
    return pSp;
}
