//
// -- LICENSE: The MIT License(MIT)
//
// Copyright(c) 2021 DinS - https://dins.site
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
#include "VideoPlayer.h"
#include <Urho3D/Resource/ResourceCache.h>
#include <Urho3D/Scene/Scene.h>
#include <Urho3D/Scene/SceneEvents.h>
#include <Urho3D/Audio/SoundSource.h>
#include <Urho3D/Urho2D/Renderer2D.h>

VideoPlayer::VideoPlayer(Context* context) : StaticSprite2D(context), _fFrameDuration(0.0f), _fElapsedTime(0.0f)
{
}

VideoPlayer::~VideoPlayer()
{
    auto strResName = _pMpg->GetName();
    _pMpg.Reset();
    auto pCache = GetSubsystem<Urho3D::ResourceCache>();
    pCache->ReleaseResource<MpgVideo>(strResName);
}

void VideoPlayer::RegisterObject(Context* context)
{
    context->RegisterFactory<VideoPlayer>("Component");
}

void VideoPlayer::SetMpg(MpgVideo* pMpgRes, float fVolume)
{
    if (_pMpg == pMpgRes)
        return;
    //1.基本信息
    _pMpg = pMpgRes;
    _fFrameDuration = 1.0f / _pMpg->GetFrameRate();    //因为update是以秒计算的，所以要将fps换算成每帧占多少秒
    _fElapsedTime = 0.0f;
    //2.显示第一帧图像
    auto pFrameImg = _pMpg->GetNextFrame();
    SetSprite(pFrameImg);
    //3.播放音轨
    auto pSS = node_->CreateComponent<Urho3D::SoundSource>();
    pSS->SetAutoRemoveMode(Urho3D::REMOVE_COMPONENT);
    pSS->Play(_pMpg->GetAudio());
    pSS->SetGain(fVolume);
    /*额外说明：为何不逐帧播放而是将音频一口气全部送入音频设备流里？
     因为音频的帧数跟视频的帧数往往不一样，如果在Update里图像一起推进，那么就会造成视频音频不同步
    解决方法也很简单：将音频全部flush然后SDL会实时播放，结果就是同步了 */
}

//这个方法的作用应该是当加入Scene后，注册E_SCENEPOSTUPDATE事件
//注册E_SCENEPOSTUPDATE事件的前提下才会触发HandleScenePostUpdate方法
void VideoPlayer::OnSceneSet(Scene* scene)
{
    StaticSprite2D::OnSceneSet(scene);
    SubscribeToEvent(scene, Urho3D::E_SCENEPOSTUPDATE, URHO3D_HANDLER(VideoPlayer, HandleScenePostUpdate));
}

//E_SCENEPOSTUPDATE事件触发，更新帧图像
void VideoPlayer::HandleScenePostUpdate(StringHash eventType, VariantMap& eventData)
{
    float timeStep = eventData[Urho3D::ScenePostUpdate::P_TIMESTEP].GetFloat();
    UpdateFrame(timeStep);
}


void VideoPlayer::UpdateFrame(float timeStep)
{
    if (_pMpg == nullptr)
        return;
    _fElapsedTime += timeStep;
    //如果当前帧的时间大于等于每帧额定时间，就推进至下一帧
    if (_fElapsedTime >= _fFrameDuration)
    {
        auto pFrameImg = _pMpg->GetNextFrame();
        if (pFrameImg.NotNull())
        {
            //auto pPreFrame = GetSprite()->GetTexture();
            SetSprite(pFrameImg);
            _fElapsedTime -= _fFrameDuration;
            //必须调整Renderer2D的代码才能够真正释放资源内存
            //GetScene()->GetComponent<Urho3D::Renderer2D>()->UncacheTexture(pPreFrame);
        }
        else
        {
            node_->RemoveComponent<VideoPlayer>();
        }
    }
}
